#--------------------------------------------------------#
#  Resolution de l'equation d'advection diffusion        #
#                                                        #
#  \partial_t u + a\partial_x u -\nu \partial_x^2 u =f   #
#  Conditions aux limites de Dirichlet a gauche,         #
#  Neumann a droite                                      #
#                                                        #
#  Resolution par decomposition de domaines              #
#  Methode de Schwarz Waveform Relaxation                #
#                                                        #
#  C. Besse, IMT Toulouse, 2018                          #
#--------------------------------------------------------#

import numpy as np
import matplotlib.pylab as plt
import scipy.sparse as sp
import scipy.sparse.linalg as slin
plt.ion()

## Donnees du probleme
a0=-5; b0=5
T=1
L=b0-a0
N=1000
dx=L/N
x=np.arange(a0,b0+dx,dx)
Niter=100
dt=T/Niter

a=1
nu=1

overlap=50

a1=a0
b1=0+overlap*dx
a2=0
b2=b0

x1=np.arange(a1,b1+dx,dx)
x2=np.arange(a2,b2+dx,dx)

## Donnee initiale
phi=np.exp(-x**2)

## Terme source
f=0.5*(x>-1)*(x<1)

## Restriction des donnees aux deux domaines
phi1=phi[0:501+overlap]
phi2=phi[500:N+1]
f1=f[0:501+overlap]
f2=f[500:N+1]


## Valeur des noeuds d'echange
N1 = phi1.size
N2 = phi2.size
p1 = N1-overlap-1
p2 = 1+overlap-1


## Assemblage des matrices
vec   = np.ones((N1),dtype=np.float64);
data  = (2/dt)*np.array([vec]); diags = np.array([0])
Mt1   = sp.spdiags(data,diags,N1,N1, format='lil')
data  = (0.5/dx)*np.array([-vec,vec]); diags = np.array([-1,1])
Ma1   = sp.spdiags(data,diags,N1,N1, format='lil')
data  = (1/dx**2)*np.array([vec,-2*vec,vec]); diags = np.array([-1,0,1])
Mlap1 = sp.spdiags(data,diags,N1,N1, format='lil')

vec   = np.ones((N2),dtype=np.float64);
data  = (2/dt)*np.array([vec]); diags = np.array([0])
Mt2   = sp.spdiags(data,diags,N2,N2, format='lil')
data  = (0.5/dx)*np.array([-vec,vec]); diags = np.array([-1,1])
Ma2   = sp.spdiags(data,diags,N2,N2, format='lil')
data  = (1/dx**2)*np.array([vec,-2*vec,vec]); diags = np.array([-1,0,1])
Mlap2 = sp.spdiags(data,diags,N2,N2, format='lil')


M1=Mt1+a*Ma1-nu*Mlap1
M2=Mt2+a*Ma2-nu*Mlap2

cl1=np.zeros(N1,dtype=np.float64)
cl2=np.zeros(N2,dtype=np.float64)

# Conditions aux limites du probleme global
# Dirichlet a gauche
gl=0
M1[0,0]=1
M1[0,1]=0
cl1[0]=gl

# Neumann a droite
gr=0
M2[N2-1,N2-1]=2/dt+2*nu/dx**2
M2[N2-1,N2-2]=-2*nu/dx**2
cl2[N2-1]=(2*nu/dx-a)*gr

# Conditions d'interface
M1[N1-1,N1-1]=1
M1[N1-1,N1-2]=0

M2[0,0]=1
M2[0,1]=0

Transm_12km1 = np.zeros(Niter,dtype=np.float64)
Transm_21km1 = np.zeros(Niter,dtype=np.float64)
Transm_12k   = np.zeros(Niter,dtype=np.float64)
Transm_21k   = np.zeros(Niter,dtype=np.float64)

M1=M1.tocsc(); solve1 = slin.factorized(M1)
M2=M2.tocsc(); solve2 = slin.factorized(M2)

fig = plt.figure(1)
plt.clf()
line1 = plt.Line2D([], [], color='blue', linewidth=1)
line11= plt.Line2D([], [], color='blue', linewidth=1, marker='*')
line2 = plt.Line2D([], [], color='red', linewidth=1)
line21= plt.Line2D([], [], color='red', linewidth=1, marker='o')
ax=plt.subplot(1,1,1)
ax.add_line(line1)
ax.add_line(line2)
ax.add_line(line11)
ax.add_line(line21)
line1.set_data(x1[:], phi1[:])
line11.set_data(x1[-1], phi1[-1])
line2.set_data(x2[:], phi2[:])
line21.set_data(x2[0], phi2[0])
ax.set_xlim([a0, b0])
ax.set_ylim([0, 1.2])
#ax.set_aspect('equal', 'datalim')
ax.set_xlabel('x') # Setting x-axis label
ax.set_ylabel('u') # Setting y-axis label
plt.draw()
plt.show()
plt.pause(0.5)

for k in range(15):
  phi1=phi[0:501+overlap]
  phi2=phi[500:N+1]

  # Resolution sur le domaine 1
  for j in range(Niter-1):
    rhs1=f1+cl1+(2/dt)*phi1 
    rhs1[0]=cl1[0]
    rhs1[N1-1]=Transm_21km1[j]
    v=solve1(rhs1)
    phi1=2*v-phi1 
    Transm_12k[j]=v[p1]

  # Resolution sur le domaine 2
  for j in range(Niter-1):
    rhs2=f2+cl2+(2/dt)*phi2 
    rhs2[0]=Transm_12km1[j]
    v=solve2(rhs2)
    phi2=2*v-phi2 
    Transm_21k[j]=v[p2]

  print('iteration ',k+1)
      
  Transm_12km1=Transm_12k.copy()
  Transm_21km1=Transm_21k.copy()
  line1.set_data(x1[:], phi1[:])
  line11.set_data(x1[-1], phi1[-1])
  line2.set_data(x2[:], phi2[:])
  line21.set_data(x2[0], phi2[0])
  plt.draw()
  plt.show()
  plt.pause(0.5)


